/*
 * Copyright (c) 1983 Regents of the University of California.
 * All rights reserved.
 *
 * Copyright 1999 David A. Holland. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notices, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notices, this list of conditions, and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgements:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 *	This product includes software developed by David A. Holland.
 * 4. Neither the name of the University nor the names of its contributors
 *    nor the names of other copyright holders may be used to endorse or 
 *    promote products derived from this software without specific prior 
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS, CONTRIBUTORS, AND OTHER AUTHORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED 
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS, CONTRIBUTORS, OR
 * OTHER AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * From: @(#)display.c	5.4 (Berkeley) 6/1/90
 *
 * Heavily modified 11/99 by dholland to add scrollback support.
 *
 * multibyte support 2002 by Gerd Knorr <kraxel@bytesex.org>
 * also got window resize without crashing for free because
 * I had to replace some fixed-length arrays by dynamically
 * allocated ones :-)
 */
char display_rcsid[] = 
  "$Id: display.c,v 1.15 2000/07/29 18:50:27 dholland Exp $";

/*
 * The window 'manager', initializes curses and handles the actual
 * displaying of text
 */
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <assert.h>
#include <wchar.h>
#include <iconv.h>
#include <curses.h>
#include <langinfo.h>
#include <sys/select.h>
#include "talk.h"

#define MAX_MAXLINES 16384

typedef struct {
	wchar_t *l_text;
} line;

typedef struct {
	line *s_lines;          /* Text in scroll buffer */
	int   s_nlines;         /* Number of valid lines in s_lines[] */
	int   s_maxlines;       /* Max already-allocated number of lines */
	int   s_scrollup;       /* Number of lines scrolled back */
	char *s_typebuf_raw;    /* Current input line */
        int   s_typebuf_len;    /* Current position in input line */
	wchar_t *s_typebuf_w;
	int   s_typebuf_wlen;
        int   s_typebufpos;     /* Current position in input line */
	char  cerase;           /* Erase-character key */
	char  werase;           /* Erase-word key */
	char  lerase;           /* Erase-line key */
} window;

static const char *topmessage;  /* Message at top of screen, if any */
static window my_win;           /* Scrolling data for top window */
static window his_win;          /* Scrolling data for bottom window */
static int sepline;             /* Line where separator is */
static int last_was_bot;        /* if 1, last win with activity was bottom */
char *peer_encoding = NULL;

static void
wait_for_data(int fd)
{
    fd_set          se;
    
    FD_ZERO(&se);
    FD_SET(fd,&se);
    select(fd+1,&se,NULL,NULL,NULL);
}


static
void
do_sigwinch(int ignore)
{
	(void)ignore;

	/* Get curses to notice the size change */
	endwin();
	refresh();

	/* and now repaint. */
	dorefresh();
}

static
void
init_window(window *win)
{
	win->s_maxlines = 32;
	win->s_nlines = 0;
	win->s_lines = malloc(win->s_maxlines * sizeof(line));
	if (!win->s_lines) p_error("Out of memory");

	win->s_scrollup = 0;

	win->s_typebufpos   = 0;
	win->s_typebuf_len  = 0;
	win->s_typebuf_raw  = NULL;
	win->s_typebuf_wlen = 64;
	win->s_typebuf_w    = malloc((win->s_typebuf_wlen+1) *
				     sizeof(wchar_t));

	win->cerase = 2;   /* ^B */
	win->werase = 23;  /* ^W */
	win->lerase = 21;  /* ^U */
}

void
real_init_display(void)
{
	struct sigaction sigac;

	/* Open curses. */
	if (initscr() == NULL) {
		printf("initscr failed: TERM is not set or set to an "
		       "unknown terminal type.\n");
		exit(1);
	}

	/* Block SIGALRM while handling SIGTSTP (curses catches SIGTSTP). */
	sigaction(SIGTSTP, NULL, &sigac);
	sigaddset(&sigac.sa_mask, SIGALRM);
	sigaction(SIGTSTP, &sigac, NULL);

	/* Grab SIGWINCH (curses cannot do this correctly on its own) */
	sigaction(SIGWINCH, NULL, &sigac);
	sigaddset(&sigac.sa_mask, SIGALRM);
	sigac.sa_handler = do_sigwinch;
	sigaction(SIGWINCH, &sigac, NULL);

	/* Set curses modes. */
	cbreak();
	noecho();
	nodelay(stdscr,1); /* otherwise get_wch() may block */
	nl();        /* force cr->lf */

	init_window(&my_win);
	init_window(&his_win);

	topmessage = NULL;
	sepline = LINES/2;

	/* Forcibly clear the screen. */
	clear();
	dorefresh();
}

void
set_my_edit_chars(int cerase, int lerase, int werase)
{
	my_win.cerase = cerase;
	my_win.lerase = lerase;
	my_win.werase = werase;
}

void
set_his_edit_chars(int cerase, int lerase, int werase)
{
	his_win.cerase = cerase;
	his_win.lerase = lerase;
	his_win.werase = werase;
}

/**************************************************************/

void 
dobeep(void)
{
	beep();
}

static
void
refresh_window(window *win, int nlines, int typeline)
{
	int i, line, topline = typeline-nlines+1;
	for (i=nlines-1; i>=0; i--) {
		move(topline+i, 0);
		line = win->s_nlines - (nlines - 1) + i - win->s_scrollup;
		if (line >=0 && line < win->s_nlines) {
			addwstr(win->s_lines[line].l_text);
		}
		else if (line==win->s_nlines) {
			addwstr(win->s_typebuf_w);
		}
		clrtoeol();
	}
}

void
dorefresh(void)
{
	int i;
	erase();

	sepline = LINES/2;

	/* Separator */
	move(sepline, 0);
	for (i=0; i<COLS; i++) addch(ACS_HLINE);

	/* Top window */
	refresh_window(&my_win, sepline, sepline-1);
	
	/* Bottom window */
	refresh_window(&his_win, LINES-sepline-1, LINES-1);

	/* Message */
	if (topmessage) {
		move(0, 0);
		addch('[');
		addstr(topmessage);
		addch(']');
	}

	/* Scroll indicators */
	if (my_win.s_scrollup) {
		move(sepline, 4);
		addstr("[upper scroll]");
	}
	if (his_win.s_scrollup) {
		move(sepline, COLS-4-14);
		addstr("[lower scroll]");
	}

	if (last_was_bot) {
		if (his_win.s_scrollup) {
			move(sepline, COLS-5);
		}
		else {
			move(LINES-1, his_win.s_typebufpos);
		}
	}
	else {
		if (my_win.s_scrollup) {
			move(sepline, 17);
		}
		else {
			move(sepline-1, my_win.s_typebufpos);
		}
	}

	refresh();
}

/**************************************************************/

static
void
doscroll(window *win, int delta)
{
	int nsc = win->s_scrollup + delta;
	if (nsc<0) nsc = 0;
	if (nsc>win->s_nlines) nsc = win->s_nlines;
	win->s_scrollup = nsc;
	dorefresh();
}

/*
 * Note: at this level we trap scrolling keys and other local phenomena.
 * Erase/word erase/line erase, newlines, and so forth get passed through,
 * sent to the other guy, and ultimately come out in display().
 */
char*
dogetch(void)
{
	static int gotesc = 0;
	int scrl = sepline-2;  /* number of lines to scroll by */
	wint_t k;
	wchar_t wc[2];
	static char mbs[16];

#if 1
	/* not implemented in older libncursesw versions */
	if (OK != get_wch(&k))
	    k = L'?';
#else
	/* buggy due to stdio buffering (cut+paste shows this nicely) */
	k = getwchar();
#endif
	if (WEOF == k)
		return NULL;

	if (!gotesc && k == 27) {
		gotesc = 1;
		return NULL;
	}

	if (gotesc && k==L'p') {          /* M-p: scroll our window up */
		doscroll(&my_win, scrl);
		gotesc = 0;
		return NULL;
	}
	else if (gotesc && k==L'n') {     /* M-n: scroll our window down */
		doscroll(&my_win, -scrl);
		gotesc = 0;
		return NULL;
	}
	else if (k==(L'p'&31)) {      /* C-p: scroll other window up */
		doscroll(&his_win, scrl);
		return NULL;
	}
	else if (k==(L'n'&31)) {      /* C-n: scroll other window down */
		doscroll(&his_win, -scrl);
		return NULL;
	}
	else if (k == '\f') {        /* C-l: reprint */
		clear();
		dorefresh();
		return NULL;
	}

	gotesc = 0;
	wc[0] = k;
	wc[1] = L'\0';
	wcstombs(mbs,wc,sizeof(mbs));
	return mbs;
}

/**************************************************************/

static
void
display_lerase(window *win)
{
	free(win->s_typebuf_raw);
	win->s_typebuf_raw  = NULL;
	win->s_typebuf_len  = 0;
	win->s_typebuf_w[0] = L'\0';
	win->s_typebufpos   = 0;
}

static
void
discard_top_line(window *win)
{
	int i;
	assert(win->s_nlines>0);
	free(win->s_lines[0].l_text);
	for (i=0; i<win->s_nlines-1; i++) {
		win->s_lines[i] = win->s_lines[i+1];
	}
	win->s_nlines--;
}

static
void
display_eol(window *win)
{
	line *tmpl;
	wchar_t *tmps;

	if (win->s_nlines == win->s_maxlines) {
		if (win->s_maxlines < MAX_MAXLINES) {
			win->s_maxlines *= 2;
			tmpl = realloc(win->s_lines, 
				       win->s_maxlines*sizeof(line));
		}
		else {
			/* Reached size limit - pretend realloc failed */
			tmpl = NULL;
		}

		if (!tmpl) {
			discard_top_line(win);
		}
		else {
			win->s_lines = tmpl;
		}
	}
	assert(win->s_nlines < win->s_maxlines);
	
	while ((tmps = wcsdup(win->s_typebuf_w))==NULL && win->s_nlines>0) {
		discard_top_line(win);
	}
	if (!tmps) {
		p_error("Out of memory");
	}

	win->s_lines[win->s_nlines++].l_text = tmps;

	display_lerase(win);

	if (win==&my_win) topmessage = NULL;
}

static wchar_t *try_convert(char *charset, char *src, size_t ilen)
{
	size_t olen;
	wchar_t *dest;
	char *d;
	iconv_t ic;
	int rc;

	ic = iconv_open("WCHAR_T", charset);
	if (NULL == ic)
		return NULL;

	dest = malloc((ilen+1) * sizeof(wchar_t));
	d = (char*)dest;
	if (NULL == dest)
		p_error("Out of memory");
	olen = ilen * sizeof(wchar_t);
	rc = iconv(ic,&src,&ilen,&d,&olen);
	iconv_close(ic);
	if (-1 == rc) {
		free(dest);
		return NULL;
	}
	*((wchar_t*)d) = L'\0';
	return dest;
}

static int convert_back(char *charset,
			char *dest, int dlen,
			wchar_t *src, int slen)
{
	size_t ilen,olen;
	char *ibuf,*obuf;
	iconv_t ic;
	
	ic = iconv_open(charset, "WCHAR_T");
	if (NULL == ic)
		return -1;

	ilen = slen * sizeof(wchar_t);
	olen = dlen;
	ibuf = (char*)src;
	obuf = dest;
	iconv(ic,&ibuf,&ilen,&obuf,&olen);
	iconv_close(ic);
	return obuf - dest;
}

static int display_addchar(window *win, int *out, int *pos, wchar_t ch)
{
	int width;

	width = wcwidth(ch);
	if (-1 == width)
		/* non-printable by the current locale */
		ch = L'?';
	width = wcwidth(ch);
	    
	if (*out >= win->s_typebuf_wlen) {
		win->s_typebuf_wlen += 16;
		win->s_typebuf_w = realloc(win->s_typebuf_w,
					   (win->s_typebuf_wlen+1) *
					   sizeof(wchar_t));
		if (NULL == win->s_typebuf_w)
			p_error("Out of memory");
	}
	if (*pos + width > COLS-1)
		return -1;
	win->s_typebuf_w[*out] = ch;
	*out += 1;
	*pos += width;
	return 0;
}

static void display_delchar(window *win, int *out, int *pos)
{
	*out -= 1;
	*pos -= wcwidth(win->s_typebuf_w[*out]);
}

/*
 * Display some text on somebody's window, processing some control
 * characters while we are at it.
 */
void
display(int hiswin, unsigned char *text, int size)
{
	static char msg[80];
	int in,out,pos;
	wchar_t *tmp;
	int tmplen;
	char *lcharset,*charset = NULL;
	window *win = hiswin ? &his_win : &my_win;
	last_was_bot = hiswin;

	/* append new data to current line */
	win->s_typebuf_raw = realloc(win->s_typebuf_raw,
				     win->s_typebuf_len + size);
	if (NULL == win->s_typebuf_raw)
		p_error("Out of memory");
	memcpy(win->s_typebuf_raw+win->s_typebuf_len,text,size);
	win->s_typebuf_len += size;

	/* convert the line to wchar_t */
	lcharset = nl_langinfo(CODESET);
	if (hiswin) {
		/*
		 * Because as talk doesn't say what charset the other
		 * end uses, we have to do some guesswork here ...
		 *
		 *   Very first try is encoding specified on the command
		 *   line option (if present).
		 *
		 *   First try is utf-8 which is easy to identify
		 *   because iconv very likely failes if the guess is
		 *   wrong (or it doesn't matter if the guess is wrong
		 *   because the received text is plain us-ascii which
		 *   in turn is a subset utf-8 ...).
		 *
		 *   Next try is the charset used by the current
		 *   locale, taking in account that the people talking
		 *   to each other likely use the same language and
		 *   thus the same locale.
		 *
		 *   If nothing else works, we use iso-8859-1 as
		 *   fallback.
		 */
		char *try[] = {
			peer_encoding,
			"UTF-8",
			lcharset,
			"ISO-8859-1",
		};
		size_t i;
		
		for (i = 0, tmp = NULL; i < sizeof(try)/sizeof(char*); i++) {
			if (NULL == try[i])
				continue;
			charset = try[i];
			tmp = try_convert(charset,win->s_typebuf_raw,
					  win->s_typebuf_len);
			if (tmp)
				break;
		}
#if 1
		/* nice for trouble shooting ... */
		sprintf(msg,"guessed peer charset: %s",charset);
		message(msg);
#endif
	} else {
		charset = lcharset;
		tmp = try_convert(charset,win->s_typebuf_raw,
				  win->s_typebuf_len);
	}
	if (NULL == tmp)
		p_error("iconv");

	/* (re-)build wchar_t line buffer */
	tmplen = wcslen(tmp);
	
	for (in = 0, out = 0, pos = 0; in < tmplen; in++) {
		if (tmp[in] == L'\n' || tmp[in]==L'\r') {
			break;
		}
		else if (tmp[in]==L'\b' || 
			 tmp[in]==127 || 
			 tmp[in]==win->cerase) {
			if (out > 0)
				display_delchar(win,&out,&pos);
		}
		else if (tmp[in] == win->werase) {
			while (out > 0 && win->s_typebuf_w[out-1] != L' ')
				display_delchar(win,&out,&pos);
		}
		else if (tmp[in] == win->lerase) {
			/* line kill */
			out = 0;
			pos = 0;
		}
		else if (tmp[in] == L'\a') {
			if (in+1 == tmplen)
				beep();
		}
		else if (tmp[in] == '\f') {
			/* nothing */
		}
		else if (tmp[in] == L'\t') {
			while ((out % 8) != 0) {
				if (-1 == display_addchar(win,&out,&pos,L' '))
					break;
			}
			if ((out % 8) != 0)
				break;
		}
		else if ((tmp[in]) < L' ') {
			if (-1 == display_addchar(win,&out,&pos,L'^'))
				break;
			if (-1 == display_addchar(win,&out,&pos,tmp[in] + 64))
				break;
		}
		else {
			if (-1 == display_addchar(win,&out,&pos,tmp[in]))
				break;
		}
	}
	win->s_typebuf_w[out] = L'\0';
	win->s_typebufpos = pos;

	if (in != tmplen) {
		/* line break */
		display_eol(win);
		if (tmp[in] == L'\n' || tmp[in]==L'\r')
			in++;
	}
	if (in != tmplen) {
		/* more data for the next line ... */
		size = convert_back(charset,text,size,tmp+in,tmplen-in);
		display(hiswin,text,size);
	} else {
		dorefresh();
	}
	free(tmp);
}

/**************************************************************/

/*
 * Display string in the standard location
 */
void
message(const char *string)
{
	topmessage = string;
	dorefresh();
}

static
int
check_alt_screen(void)
{
	const char *rmcup, *smcup;
	rmcup = tigetstr("rmcup");
	smcup = tigetstr("smcup");
	return rmcup || smcup;
}

static
void
middle_message(const char *string, const char *string2)
{
	int line = sepline;
	int width, start;

	dorefresh();

	width = 2 + strlen(string) + strlen(string2) + 2;
	start = (COLS - width)/2;
	if (start<0) start = 0;

	move(line, start);
	addstr("[ ");
	addstr(string);
	addstr(string2);
	addstr(" ]");
	refresh();
}

static
void
do_quit(void)
{
	move(LINES-1, 0);
	clrtoeol();
	refresh();
	endwin();

	if (invitation_waiting) {
		send_delete();
	}
	exit(0);
}

/*
 * p_error prints the system error message on the standard location
 * on the screen and then exits. (i.e. a curses version of perror)
 */
void
p_error(const char *string) 
{
	char msgbuf[256];
	int hold = check_alt_screen();

	snprintf(msgbuf, sizeof(msgbuf), "%s: %s", string, strerror(errno));
	middle_message(msgbuf, hold ? ". Press any key..." : "");

	if (hold) {
		/* alternative screen, wait before exiting */
		wait_for_data(0);
		getch();
	}

	do_quit();
}

/*
 * All done talking...hang up the phone and reset terminal thingy
 */
void
quit(int direct)
{
	int hold = check_alt_screen();

	if (direct != 1 && hold) {
		/* alternative screen, prompt before exit */
		middle_message("Press any key to exit", "");
		wait_for_data(0);
		getch();
	}
	else {
		/* Make sure any message appears */
		dorefresh();
	}

	do_quit();
}
